/*******************************************************************************
 * Copyright 2020 Graz University of Technology
 * MOA ZS has been developed in a cooperation between EGIZ
 * and Graz University of Technology.
 * Licensed under the EUPL, Version 1.2 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * https://joinup.ec.europa.eu/news/understanding-eupl-v12
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 *******************************************************************************/
package at.gv.egiz.moazs;

import at.gv.egiz.moazs.scheme.SOAPUtils;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import java.nio.charset.StandardCharsets;

import static org.assertj.core.api.Assertions.assertThat;

/**
 * @author Christof Rabensteiner
 *
 */
public class SOAPUtilsTest {

    SOAPUtils utils;

    private static final String SOAP_MESSAGE =
            "<soap:Envelope xmlns:soap=\"http://www.w3.org/2003/05/soap-envelope\"><soap:Body>" +
            "<DeliveryResponse xmlns=\"http://reference.e-government.gv.at/namespace/zustellung/msg/phas" +
            "e2/20181206#\" xmlns:ns2=\"http://reference.e-government.gv.at/namespace/persondata/phase2/" +
            "20181206#\" xmlns:ns3=\"http://www.w3.org/2000/09/xmldsig#\"><PartialSuccess><DeliverySyste" +
            "m>https://testzustellsystem.egiz.gv.at</DeliverySystem><ZSDeliveryID>zs-valid-delivery-requ" +
            "est-id</ZSDeliveryID><AppDeliveryID>valid-delivery-request-id</AppDeliveryID><GZ>12345</GZ>" +
            "</PartialSuccess></DeliveryResponse></soap:Body></soap:Envelope>";


    public static final String FORMATTED_MESSAGE =
            "<soapenv:Envelope\n" +
            "        xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\"\n" +
            "        xmlns:msg=\"http://reference.e-government.gv.at/namespace/zustellung/msg/phase2/20181206#\">\n" +
            "    <soapenv:Body>\n" +
            "        <msg:DeliveryNotification>\n" +
            "            <msg:AppDeliveryID>formatted-adid</msg:AppDeliveryID>\n" +
            "        </msg:DeliveryNotification>\n" +
            "    </soapenv:Body>\n" +
            "</soapenv:Envelope>\n";

    public static final String CLUTTERED_MESSAGE =
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
            "<soapenv:Envelope " +
            "  xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\"\n " +
            "  xmlns:msg=\"http://reference.e-government.gv.at/namespace/zustellung/msg/phase2/20181206#\">\n" +
            "    <soapenv:Body unexpected-attribute=\"unexpectedvalue\">  \n" +
            "        <msg:DeliveryNotification>\n" +
            "            <msg:AppDeliveryID attribute=\"some-value\"><element-that-got-lost />cluttered-adid</msg:AppDeliveryID>\n" +
            "        </msg:DeliveryNotification>\n" +
            "    </soapenv:Body>\n" +
            "</soapenv:Envelope>\n";

    private static final String MISSING_BODY_MESSAGE =
            "<soap:Envelope xmlns:soap=\"http://www.w3.org/2003/05/soap-envelope\"><NotABodyTag></NotABodyTag></soap:Envelope>";

    private static final String EMPTY_BODY_MESSAGE =
            "<soap:Envelope xmlns:soap=\"http://www.w3.org/2003/05/soap-envelope\"><soap:Body>       </soap:Body></soap:Envelope>";

    public static final String EMPTY_APP_DELIVERY_ID_MESSAGE =
            "<soapenv:Envelope\n" +
                    "        xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\"\n" +
                    "        xmlns:msg=\"http://reference.e-government.gv.at/namespace/zustellung/msg/phase2/20181206#\">\n" +
                    "    <soapenv:Body>\n" +
                    "        <msg:DeliveryNotification>\n" +
                    "            <msg:AppDeliveryID />\n" +
                    "        </msg:DeliveryNotification>\n" +
                    "    </soapenv:Body>\n" +
                    "</soapenv:Envelope>\n";

    public static final String BLANK_APP_DELIVERY_ID_MESSAGE =
            "<soapenv:Envelope\n" +
                    "        xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\"\n" +
                    "        xmlns:msg=\"http://reference.e-government.gv.at/namespace/zustellung/msg/phase2/20181206#\">\n" +
                    "    <soapenv:Body>\n" +
                    "        <msg:DeliveryNotification>\n" +
                    "            <msg:AppDeliveryID>   </msg:AppDeliveryID>\n" +
                    "        </msg:DeliveryNotification>\n" +
                    "    </soapenv:Body>\n" +
                    "</soapenv:Envelope>\n";

    @Before
    public void setup() {
        utils = new SOAPUtils();
    }

    @Test
    public void canParseSoapMessage() throws Exception {
        parseAndCheckLocalName(SOAP_MESSAGE);
    }

    @Test
    public void canParseFormattedSoapMessage() throws Exception {
        parseAndCheckLocalName(FORMATTED_MESSAGE);
    }

    @Test
    public void canParseClutteredSoapMessage() throws Exception  {
        parseAndCheckLocalName(CLUTTERED_MESSAGE);
    }

    private void parseAndCheckLocalName(String text) throws Exception {
        byte[] bytes = text.getBytes(StandardCharsets.UTF_8);
        Element root = utils.toDOM(bytes);
        assertThat(root.getLocalName()).isEqualTo("Envelope");
    }

    @Test
    public void canUnwrapSoapMessage() throws Exception {
        getFirstChildAndCheckLocalName(SOAP_MESSAGE, "DeliveryResponse");
    }

    @Test
    public void canUnwrapFormattedSoapMessage() throws Exception {
        getFirstChildAndCheckLocalName(FORMATTED_MESSAGE, "DeliveryNotification");
    }

    @Test
    public void canUnwrapClutteredSoapMessage() throws Exception {
        getFirstChildAndCheckLocalName(CLUTTERED_MESSAGE, "DeliveryNotification");
    }

    @Test(expected=MoaZSException.class)
    public void failUnwrapMissingBody() throws Exception {
        getFirstChildAndCheckLocalName(MISSING_BODY_MESSAGE, "");
    }

    @Test(expected=MoaZSException.class)
    public void failUnwrapEmptyBody() throws Exception {
        getFirstChildAndCheckLocalName(EMPTY_BODY_MESSAGE, "");
    }

    private void getFirstChildAndCheckLocalName(String message, String expectedLocalNameOfChild) throws Exception {
        byte[] bytes = message.getBytes(StandardCharsets.UTF_8);
        Element soapRoot = utils.toDOM(bytes);
        Node child = utils.getChildElementOfSoapBody(soapRoot);
        var actualLocalName = child.getLocalName();
        assertThat(actualLocalName).isEqualTo(expectedLocalNameOfChild);
    }

    @Test
    public void getAppDeliveryIDFromSoapMessage() throws Exception {
        getAppDeliveryIDAndCheck(SOAP_MESSAGE, "valid-delivery-request-id");
    }

    @Test
    public void getAppDeliveryIDFromFormattedMessage() throws Exception {
        getAppDeliveryIDAndCheck(FORMATTED_MESSAGE, "formatted-adid");
    }

    @Test
    public void getAppDeliveryIDFromClutteredMessage() throws Exception {
        getAppDeliveryIDAndCheck(CLUTTERED_MESSAGE, "cluttered-adid");
    }

    @Test(expected = MoaZSException.class)
    public void failToRetrieveAppDeliveryIDWhenBodyIsMissing() throws Exception {
        getAppDeliveryIDAndCheck(MISSING_BODY_MESSAGE, "");
    }

    @Test(expected = MoaZSException.class)
    public void failToRetrieveEmptyAppDeliveryID() throws Exception {
        getAppDeliveryIDAndCheck(EMPTY_APP_DELIVERY_ID_MESSAGE, "");
    }

    @Test(expected = MoaZSException.class)
    public void failToRetrieveBlankAppDeliveryID() throws Exception {
        getAppDeliveryIDAndCheck(BLANK_APP_DELIVERY_ID_MESSAGE, "");
    }

    private void getAppDeliveryIDAndCheck(String message, String expectedValue) throws Exception {
        byte[] bytes = message.getBytes(StandardCharsets.UTF_8);
        Element soapRoot = utils.toDOM(bytes);
        String actualAppDeliveryID = utils.getAppDeliveryIDFrom(soapRoot);
        assertThat(actualAppDeliveryID).isEqualTo(expectedValue);
    }

}
